@@ -2,7 +2,7 @@
 
 mod connection;
 
-use mpd_protocol::Response as RawResponse;
+use mpd_protocol::{AsyncConnection, Response as RawResponse};
 use tokio::{
     io::{AsyncRead, AsyncWrite},
     sync::{
@@ -10,14 +10,12 @@ use tokio::{
         oneshot,
     },
 };
-use tracing::{debug, span, trace, warn, Level, Span};
-use tracing_futures::Instrument;
+use tracing::{debug, error, span, trace, warn, Instrument, Level};
 
-use std::fmt::Debug;
-use std::sync::{
-    atomic::{AtomicUsize, Ordering},
-    Arc,
-};
+use std::error::Error;
+use std::fmt;
+use std::io;
+use std::sync::Arc;
 
 use crate::commands::{self as cmds, responses::Response, Command, CommandList};
 use crate::errors::CommandError;
@@ -26,14 +24,11 @@ use crate::state_changes::StateChanges;
 
 type CommandResponder = oneshot::Sender<Result<RawResponse, CommandError>>;
 
-/// Counter for connection IDs for logging.
-static CONNECTION_COUNTER: AtomicUsize = AtomicUsize::new(0);
-
-/// Result returned by a connection attempt.
+/// Components of a connection.
 ///
-/// If successful, this contains a [`Client`] which you can use to issue commands, and a
-/// [`StateChanges`] value, which is a stream that receives notifications from the server.
-pub type ConnectResult = Result<(Client, StateChanges), MpdProtocolError>;
+/// This contains a [`Client`] which you can use to issue commands, and a [`StateChanges`] value,
+/// which is a stream that receives notifications from the server.
+pub type Connection = (Client, StateChanges);
 
 /// A client connected to an MPD instance.
 ///
@@ -47,7 +42,6 @@ pub type ConnectResult = Result<(Client, StateChanges), MpdProtocolError>;
 pub struct Client {
     commands_sender: Sender<(RawCommandList, CommandResponder)>,
     protocol_version: Arc<str>,
-    span: Span,
 }
 
 impl Client {
@@ -63,33 +57,37 @@ impl Client {
     /// # Errors
     ///
     /// This will return an error if sending the initial commands over the given transport fails.
-    pub async fn connect<C>(connection: C) -> ConnectResult
+    pub async fn connect<C>(connection: C) -> Result<Connection, MpdProtocolError>
     where
         C: AsyncRead + AsyncWrite + Unpin + Send + 'static,
     {
-        let id = CONNECTION_COUNTER.fetch_add(1, Ordering::Relaxed);
-        let span = span!(Level::DEBUG, "client connection", id);
+        do_connect(connection, None).await.map_err(|e| match e {
+            ConnectWithPasswordError::ProtocolError(e) => e,
+            ConnectWithPasswordError::IncorrectPassword => unreachable!(),
+        })
+    }
 
-        let (state_changes_sender, state_changes) = mpsc::unbounded_channel();
-        let (commands_sender, commands_receiver) = mpsc::channel(1);
-
-        let connection = connection::connect(connection, &span).await?;
-
-        let protocol_version = Arc::from(connection.codec().protocol_version());
-
-        tokio::spawn(
-            connection::run_loop(connection, commands_receiver, state_changes_sender)
-                .instrument(span!(parent: &span, Level::TRACE, "run loop")),
-        );
-
-        let state_changes = StateChanges { rx: state_changes };
-        let client = Client {
-            commands_sender,
-            protocol_version,
-            span,
-        };
-
-        Ok((client, state_changes))
+    /// Connect to the password-protected MPD server using the given connection and password.
+    ///
+    /// Commonly used with [TCP connections](tokio::net::TcpStream) or [Unix
+    /// sockets](tokio::net::UnixStream).
+    ///
+    /// # Panics
+    ///
+    /// Since this spawns a task internally, this will panic when called outside a Tokio runtime.
+    ///
+    /// # Errors
+    ///
+    /// This will return an error if sending the initial commands over the given transport fails,
+    /// or if the password is incorrect.
+    pub async fn connect_with_password<C>(
+        connection: C,
+        password: &str,
+    ) -> Result<Connection, ConnectWithPasswordError>
+    where
+        C: AsyncRead + AsyncWrite + Unpin + Send + 'static,
+    {
+        do_connect(connection, Some(password)).await
     }
 
     /// Send a [command].
@@ -130,25 +128,6 @@ impl Client {
         <L as CommandList>::parse_responses(frames).map_err(Into::into)
     }
 
-    /// Send the connection password, if necessary.
-    ///
-    /// If the MPD instance the `Client` is connected to is protected by a password, sending
-    /// commands may result in "Permission denied" errors prior to sending the correct password.
-    ///
-    /// This will not send anything if `None` is passed.
-    ///
-    /// # Errors
-    ///
-    /// Sending an incorrect password will result in a [`CommandError::ErrorResponse`], any other
-    /// errors are returned in the same conditions as [`Client::raw_command`].
-    pub async fn password(&self, password: Option<String>) -> Result<(), CommandError> {
-        if let Some(pw) = password {
-            self.command(cmds::Password(pw)).await
-        } else {
-            Ok(())
-        }
-    }
-
     /// Send the given command, and return the response to it.
     ///
     /// # Errors
@@ -170,7 +149,7 @@ impl Client {
     /// *any* of the commands in the list return an error condition, the entire list will be
     /// treated as an error.
     ///
-    /// You may recover possible succesful fields in a response from the [error].
+    /// You may recover possible successful fields in a response from the [error].
     ///
     /// [error]: crate::errors::CommandError::ErrorResponse
     pub async fn raw_command_list(
@@ -211,7 +190,7 @@ impl Client {
     ///
     /// # Return value
     ///
-    /// If this method returns succesfully, a return value of  `None` indicates that no album art
+    /// If this method returns successfully, a return value of  `None` indicates that no album art
     /// for the given URI was found. Otherwise, you will get a tuple consisting of the raw binary
     /// data, and an optional string value that contains a MIME type for the data, if one was
     /// provided by the server.
@@ -263,7 +242,7 @@ impl Client {
                 out.extend_from_slice(resp.data());
                 debug!(length = expected_size, "found separate file album art");
             } else {
-                debug!("no embedded or separte album art found");
+                debug!("no embedded or separate album art found");
                 return Ok(None);
             }
         }
@@ -306,11 +285,117 @@ impl Client {
     }
 }
 
+/// Perform the initial handshake to the server.
+async fn do_connect<IO: AsyncRead + AsyncWrite + Unpin + Send + 'static>(
+    io: IO,
+    password: Option<&str>,
+) -> Result<Connection, ConnectWithPasswordError> {
+    let span = span!(Level::DEBUG, "client connection");
+
+    let (state_changes_sender, state_changes) = mpsc::unbounded_channel();
+    let (commands_sender, commands_receiver) = mpsc::channel(1);
+
+    let mut connection = match AsyncConnection::connect(io).instrument(span.clone()).await {
+        Ok(c) => c,
+        Err(e) => {
+            error!(error = ?e, "failed to perform initial handshake");
+            return Err(e.into());
+        }
+    };
+
+    let protocol_version = Arc::from(connection.protocol_version());
+
+    if let Some(password) = password {
+        trace!(parent: &span, "sending password");
+
+        if let Err(e) = connection
+            .send(RawCommand::new("password").argument(password.to_owned()))
+            .instrument(span.clone())
+            .await
+        {
+            error!(parent: &span, error = ?e, "failed to send password");
+            return Err(e.into());
+        }
+
+        match connection.receive().instrument(span.clone()).await {
+            Err(e) => {
+                error!(parent: &span, error = ?e, "failed to receive reply to password");
+                return Err(e.into());
+            }
+            Ok(None) => {
+                error!(
+                    parent: &span,
+                    "unexpected end of stream after sending password"
+                );
+                return Err(MpdProtocolError::Io(io::Error::new(
+                    io::ErrorKind::UnexpectedEof,
+                    "connection closed while waiting for reply to password",
+                ))
+                .into());
+            }
+            Ok(Some(response)) if response.is_error() => {
+                error!(parent: &span, "incorrect password");
+                return Err(ConnectWithPasswordError::IncorrectPassword);
+            }
+            Ok(Some(_)) => {
+                trace!(parent: &span, "password accepted");
+            }
+        }
+    }
+
+    tokio::spawn(
+        connection::run_loop(connection, commands_receiver, state_changes_sender)
+            .instrument(span!(parent: &span, Level::TRACE, "run loop")),
+    );
+
+    let state_changes = StateChanges { rx: state_changes };
+    let client = Client {
+        commands_sender,
+        protocol_version,
+    };
+
+    Ok((client, state_changes))
+}
+
+/// Error returned when [connecting with a password][Client::connect_with_password] fails.
+#[derive(Debug)]
+pub enum ConnectWithPasswordError {
+    /// The provided password was not accepted by the server.
+    IncorrectPassword,
+    /// An unrelated protocol error occurred.
+    ProtocolError(MpdProtocolError),
+}
+
+impl fmt::Display for ConnectWithPasswordError {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        match self {
+            ConnectWithPasswordError::IncorrectPassword => write!(f, "incorrect password"),
+            ConnectWithPasswordError::ProtocolError(_) => write!(f, "protocol error"),
+        }
+    }
+}
+
+impl Error for ConnectWithPasswordError {
+    fn source(&self) -> Option<&(dyn Error + 'static)> {
+        match self {
+            ConnectWithPasswordError::ProtocolError(e) => Some(e),
+            ConnectWithPasswordError::IncorrectPassword => None,
+        }
+    }
+}
+
+#[doc(hidden)]
+impl From<MpdProtocolError> for ConnectWithPasswordError {
+    fn from(e: MpdProtocolError) -> Self {
+        ConnectWithPasswordError::ProtocolError(e)
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
     use crate::state_changes::Subsystem;
-    use futures::StreamExt;
+    use futures_util::StreamExt;
     use tokio_test::{assert_ok, io::Builder as MockBuilder};
 
     static GREETING: &[u8] = b"OK MPD 0.21.11\n";
